
/*
 * XKOBO, a video-oriented game
 * Copyright (C) 1995,1996  Akira Higuchi
 *     a-higuti@math.hokudai.ac.jp
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <btron/libapp.h>

#include "tiff.h"
#include "xkobo.h"

extern CSPEC cspec_system;

/********    tiny TIFF loader  ( 256 colors, not compressed )   ********/

static int pos_image   = 0;
static int pos_cmap    = 0;
static int sizex       = 0;
static int sizey       = 0;
static int endianness;

static unsigned int get_16bits(unsigned char *p)
{
    if (endianness == 0x49)
        return (p[0] + (p[1]<<8));
    return (p[1] + (p[0]<<8));
}

static unsigned int get_32bits(unsigned char *p)
{
    if (endianness == 0x49)
        return (p[0] + (p[1]<<8) + (p[2]<<16) + (p[3]<<24));
    return (p[3] + (p[2]<<8) + (p[1]<<16) + (p[0]<<24));
}

static int read_tag(unsigned char *data)
{
    unsigned char *p = data;
    int dir_p = 0;
    int tag_max;
    int tag_type;
    int i;
    
    endianness = p[0];
    if ((endianness != 0x49) && (endianness != 0x4d)) return 1;
    if (endianness != p[1]) return 2;
    if (get_16bits(p + 2) != 0x2a) return 3;
    p += 4;

    for (;;){
        dir_p = get_32bits(p);
        if (dir_p == 0) break;
        p = data + dir_p;
        
        tag_max = get_16bits(p);
        p += 2;
        for (i=0; i<tag_max; i++){
            tag_type = get_16bits(p);
            switch (tag_type){
              case 0x0100:
                {
                    /* image width */
                    sizex = get_16bits(p + 8);
                    break;
                };
              case 0x0101:
                {
                    /* image length */
                    sizey = get_16bits(p + 8);
                    break;
                };
              case 0x0102:
                {
                    /* bitss per sample */
                    if (get_16bits(p + 8) != 8)
                        return 4;
                    break;
                }
              case 0x0103:
                {
                    /* compression */
                    if (get_16bits(p + 8) != 1)
                        return 5;
                    break;
                }
              case 0x0111:
                {
                    /* strip offset */
                    pos_image = get_32bits(p + 8);
                    break;
                }
              case 0x0140:
                {
                    /* color map */
                    pos_cmap = get_32bits(p + 8);
                    break;
                }
            }
            p += 12;
        }
    }
    
    if ((!pos_image) || (!pos_cmap) || (!sizex) || (!sizey))
        return 6;
    return 0;
}

int link_tiff(unsigned char *data, win_cmap& wcm, win_chip& wch)
{
    unsigned char r[256],g[256],b[256];
    COLOR pixel[256],mask[256];
    
    if (read_tag(data)){
         printf("tiff.cc : fatal error (Illegal format)\n");
         return -1;
    }
    int i, j=0;
    for (i=0; i<256; i++){
        r[j] = *(data + pos_cmap + j*2      );
        g[j] = *(data + pos_cmap + j*2 +512 );
        b[j] = *(data + pos_cmap + j*2 +1024);
        if ((r[j]==0xff)&&(g[j]==0)&&(b[j]==0)){
            /* through */
            pixel[j] = 0x10000000;
            mask[j] = 0x10ffffff;
        }
        else {
            pixel[j] = 0x10000000 | (r[j] << 16) | (g[j] << 8) | b[j];
            mask[j] = 0x10000000;
        }
        /* shuffle */
        j = (j + 169) & 255;
    }

    GID gc;
    unsigned char *ptr;
    int x, y;
    PAT filpat;
    filpat.spat.kind = 0;
    filpat.spat.hsize = filpat.spat.vsize = 1;
    filpat.spat.bgcol = 0x10000000;
    filpat.spat.mask = FILL100;
    ptr = (unsigned char *)(data + pos_image);
    gc = gopn_mem(NULL, wch.chip, (B*)&cspec_system);
    for (y=0; y<sizey; y++){
        for (x=0; x<sizex; x++){
            RECT r; 
            r.c.left = x<<scale_log2;
            r.c.top = y<<scale_log2;
            r.c.right = r.c.left + (1<<scale_log2);
            r.c.bottom = r.c.top + (1<<scale_log2);
            filpat.spat.fgcol = pixel[*ptr];
            //            printf("(%08x)\n", pixel[*ptr]);
            gfil_rec(gc, r, &filpat, 0, G_STORE);
            ptr++;
        }
    }
    gcls_env(gc);
    ptr = (unsigned char *)(data + pos_image);
    gc = gopn_mem(NULL, wch.chip_mask, (B*)&cspec_system);    
    for (y=0; y<sizey; y++){
        for (x=0; x<sizex; x++){
            RECT r; 
            r.c.left = x<<scale_log2;
            r.c.top = y<<scale_log2;
            r.c.right = r.c.left + (1<<scale_log2);
            r.c.bottom = r.c.top + (1<<scale_log2);
            filpat.spat.fgcol = mask[*ptr];
            gfil_rec(gc, r, &filpat, 0, G_STORE);
            ptr++;
        }
    }
    gcls_env(gc);
    wcm.cmset();
    return 0;
}

